from robobosim.processors.AbstractProcessor import AbstractProcessor
from robobosim.utils.Message import Message

class LocationProcessor(AbstractProcessor):
    def __init__(self, state):
        super().__init__(state)

        self.locationCallback = None

        self.callbacklocks = {"location": False }
        self.callbacks = {"location": None }
        self.supportedMessages = ["SIM-LOCATION"]

    def process(self, status):
        name = status["name"]
        value = status["value"]

        if (name == "SIM-LOCATION"):  #
            robot_id = int(value["id"])
            if not robot_id in self.state.locations.keys():
                self.state.locations[robot_id] = {}
            self.state.locations[robot_id] = {
                "position" : {
                    "x" : float(value["tx"]),
                    "y" : float(value["ty"]),
                    "z" : float(value["tz"])
                },

                "rotation" : {
                    "x" : float(value["rx"]),
                    "y" : float(value["ry"]),
                    "z" : float(value["rz"])
                }
            }
            self.runCallback("location")
    
    def setRobotLocation(self, robot_id, position, rotation):
        name = "SIM-LOCATION-SET"
        values = {
            "id" : robot_id,
            "position" : {},
            "rotation" : {}
        }

        if (not position is None):
            values["position"]["x"] = position["x"]
            values["position"]["y"] = position["y"]
            values["position"]["z"] = position["z"]

        if (not rotation is None):
            x,y,z = rotation
            values["rotation"]["x"] = rotation["x"]
            values["rotation"]["y"] = rotation["y"]
            values["rotation"]["z"] = rotation["z"]
    
        id = self.state.getId()
        return Message(name, values, id)




from robobosim.processors.AbstractProcessor import AbstractProcessor
from robobosim.utils.Message import Message

class AGVProcessor(AbstractProcessor):
    def __init__(self, state):
        super().__init__(state)

        self.loadedCallback = None

        self.callbacklocks = {"loaded": False }
        self.callbacks = {"loaded": None }

        self.supportedMessages = ["SIM-LOADED-ITEM"]

    def process(self, status):
        name = status["name"]
        value = status["value"]

        if (name == "SIM-LOADED-ITEM"):  #
            robot_id = int(value["id"])
            if not robot_id in self.state.loaded.keys():
                self.state.loaded[robot_id] = {}
            loaded = (value["loaded"].lower() == "true")
            self.state.loaded[robot_id] = {
                "loaded" : loaded
            }
            self.runCallback("loaded")

    def loadItem(self, robot_id):
        name = "SIM-LOAD-ITEM"
        values = {
            "id" : robot_id
        }
        id = self.state.getId()
        return Message(name, values, id)
    
    def unloadItem(self, robot_id):
        name = "SIM-UNLOAD-ITEM"
        values = {
            "id" : robot_id
        }
        id = self.state.getId()
        return Message(name, values, id)



from robobosim.processors.AbstractProcessor import AbstractProcessor
from robobosim.utils.Message import Message

class ControlProcessor(AbstractProcessor):
    def __init__(self, state):
        super().__init__(state)

    def process(self, status):
        name = status["name"]
        value = status["value"]

    def resetSimulation(self):
        name = "RESET-SIM"
        values = {}
        id = self.state.getId()
        return Message(name, values, id)



import json
import websocket
import threading
import time
import sys

from robobosim.State import State
from robobosim.utils.ConnectionState import ConnectionState

from robobosim.processors.ControlProcessor import ControlProcessor
from robobosim.processors.LocationProcessor import LocationProcessor
from robobosim.processors.AGVProcessor import AGVProcessor

class Remote:
    def __init__(self, ip):
        self.ip = ip
        self.ws = None
        self.password = "REMOTE-SIM"
        self.state = State()
        self.processors = {"CONTROL": ControlProcessor(self.state),
                           "LOCATION": LocationProcessor(self.state),
                           "AGV": AGVProcessor(self.state)}

        self.wsDaemon = None
        self.connectionState = ConnectionState.DISCONNECTED

        self.timeout = 3

    def disconnect(self):
        self.ws.close()
        self.connectionState = ConnectionState.DISCONNECTED

    def wsStartup(self):
        def on_open(ws):
            print("### connection established ###")
            ws.send("PASSWORD: " + self.password)
            self.connectionState = ConnectionState.CONNECTED

        def on_message(ws, message):
            self.processMessage(message)

        def on_error(ws, error):
            print(f"ERROR: {error}")
            self.connectionState = ConnectionState.ERROR

        def on_close(ws, status_code, msg):
            self.connectionState = ConnectionState.DISCONNECTED
            if (status_code != None):
                print(f"### closed connection [{status_code}] {msg} ###")
            else:
                print("### closed connection ###")

        self.ws = websocket.WebSocketApp('ws://' + self.ip + ":50505",
                                         on_message=on_message,
                                         on_error=on_error,
                                         on_close=on_close)

        self.ws.on_open = on_open

        def runWS():
            self.ws.run_forever()

        self.wsDaemon = threading.Thread(target=runWS, name='wsDaemon')
        self.wsDaemon.setDaemon(True)
        self.wsDaemon.start()

        print("Connecting to SIM")
        self.connectionState = ConnectionState.CONNECTING

        while self.connectionState == ConnectionState.CONNECTING:
            print("wait")
            time.sleep(0.5)

    def processMessage(self, msg):
        status = json.loads(msg)
        name = status["name"]
        value = status["value"]
        processed = False
        for key in self.processors.keys():
            if self.processors[key].canProcess(name):
                self.processors[key].process(status)
                processed = True
                break

    def sendMessage(self, msg):
        if self.connectionState == ConnectionState.CONNECTED:
            self.ws.send(msg.encode())
        else:
            sys.exit("\nError: Establish connection before sending a message")

    def resetSimulation(self):
        msg = self.processors["CONTROL"].resetSimulation()
        self.sendMessage(msg)
    
    def getRobotLocation(self, robot_id):
        try:
            return self.state.locations[robot_id]
        except KeyError as e:
            return None
    
    def setRobotLocation(self, robot_id, position, rotation):
        try:
            msg = self.processors["LOCATION"].setRobotLocation(robot_id, position, rotation)
            self.sendMessage(msg)
        except KeyError as e:
            pass
    
    def loadItem(self, robot_id):
        try:
            msg = self.processors["AGV"].loadItem(robot_id)
            self.sendMessage(msg)
        except KeyError as e:
            pass
    
    def unloadItem(self, robot_id):
        try:
            msg = self.processors["AGV"].unloadItem(robot_id)
            self.sendMessage(msg)
        except KeyError as e:
            pass
    
    def isRobotLoaded(self, robot_id):
        try:
            return self.state.loaded[robot_id]["loaded"]
        except KeyError as e:
            return None

    def setLocationCallback(self, callback):
        self.processors["LOCATION"].callbacks["location"] = callback
    
    def setLoadedCallback(self, callback):
        self.processors["AGV"].callbacks["loaded"] = callback





class Message:
    """
    This class encapsulates messages to be sent to the Robobo
    Attributes:
        - name: Name of the message.
        - parameters: Dictionary containing the parameters contained in the message.
        - id: Message identifier.
    """
    def __init__(self, name, parameters, id):
        self.name = name
        self.parameters = parameters
        self.id = id

    def encode(self):
        """
        Encodes the message in JSON format.

        :return: Encoded message
        :rtype: str
        """
        st = "{\"name\":"
        st += "\"" + self.name + "\","

        st += "\"parameters\":{"

        for key in self.parameters.keys():
            st += "\"" + key + "\":" + "\"" + str(self.parameters[key]) + "\","

        if st[-1] !="{":
            st = st[:-1]

        st += "}, \"id\":\"" + str(self.id) + "\"}"

        return st



import time
from robobosim.remotelib import Remote

class RoboboSim:
    def __init__(self, ip):
        """
        Creates a new Robobo Sim library instance.

        :param ip: The IP address of the machine running RoboboSim.

        :type ip: string
        """

        self.rem = Remote(ip)

    def connect(self):
        """
        Establishes a remote connection with the Robobo Sim indicated by the IP address associated to this instance.
        """

        self.rem.wsStartup()

    def disconnect(self):
        """
        Disconnects the library from the Robobo Sim.
        """

        self.rem.disconnect()

    def wait(self, seconds):
        """
        Pauses the program for the specified time. After that time, next instruction is executed.

        :param seconds: Time to wait in seconds (>0). Decimals like 0.2 are allowed.

        :type seconds: float
        """

        time.sleep(seconds)

    def resetSimulation(self):
        """
        Resets the state of the current simulation running on RoboboSim
        """

        self.rem.resetSimulation()
    
    def getRobotLocation(self, robot_id):
        """
        Returns the location in world coordinates of the Robot specified by the index

        :param robot_id: The ID of the specified robot. Incremental, starting by 0.

        :type robot_id: int
        """

        return self.rem.getRobotLocation(robot_id)
    
    def setRobotLocation(self, robot_id, position=None, rotation=None):
        """
        Sets the location in world coordinates of the Robot specified by the index

        :param robot_id: The ID of the specified robot. Incremental, starting by 0.

        :type robot_id: int

        :param position: Optional. Dict (x,y,z) of the target global position for the robot. If not specified robot will retain position.

        :type position: dict

        :param rotation: Optional. Dict (x,y,z) of the target global rotation of the robot. If not specified robot will retain rotation.

        :type rotation: dict
        """

        self.rem.setRobotLocation(robot_id, position, rotation)

    def onNewLocation(self, callback):
        """
        Configures the callback that is called when location data is received.

        :param callback: The callback function to be called.

        :type callback: fun
        """

        self.rem.setLocationCallback(callback)

    def loadItem(self, robot_id):
        """
        If the AGV Robobo is placed on a proper load zone, it will load and carry the item provided by that zone on top of the robot
        """

        self.rem.loadItem(robot_id)
    
    def unloadItem(self, robot_id):
        """
        If the AGV Robobo is placed on a proper unload zone and is carrying a loaded item, it will drop the loaded item
        """

        self.rem.unloadItem(robot_id)
    
    def isRobotLoaded(self, robot_id):
        """
        Returns the boolean load status of the Robot specified by the index

        :param robot_id: The ID of the specified robot. Incremental, starting by 0.

        :type robot_id: int
        """

        return self.rem.isRobotLoaded(robot_id)
    
    def onNewLoaded(self, callback):
        """
        Configures the callback that is called when load status data is received.

        :param callback: The callback function to be called.

        :type callback: fun
        """

        self.rem.setLoadedCallback(callback)



import threading

class AbstractProcessor:
    def __init__(self, state):
        self.state = state
        self.supportedMessages = []
        self.callbacks = {}
        self.callbacklocks = {}

    def canProcess(self, msg):
        return msg in self.supportedMessages

    def process(self, status):
        name = status["name"]
        value = status["value"]

    def _wrapCallback(self,callback, callbackName):
        self.callbacklocks[callbackName] = True
        callback()
        self.callbacklocks[callbackName] = False

    def runCallback(self, callbackName):
        callbackExists =  self.callbacks[callbackName] is not None

        if callbackExists :

            if not self.callbacklocks[callbackName]:
                t = threading.Thread(target=self._wrapCallback, args=[self.callbacks[callbackName], callbackName])
                t.start()
            else:

                print("Too much concurrent calls to " + callbackName + "callback, ignoring")